@@ -20,41 +20,55 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
-import re
-import operator
 from collections import namedtuple
-from inspect import getargspec
-from picard.metadata import Metadata
-from picard.metadata import MULTI_VALUED_JOINER
+import datetime
+from functools import reduce
+from inspect import getfullargspec
+import operator
+import re
+import unicodedata
+
+from picard import config
+from picard.metadata import (
+    MULTI_VALUED_JOINER,
+    Metadata,
+)
 from picard.plugin import ExtensionPoint
+from picard.util import uniqify
 
 
 class ScriptError(Exception):
     pass
 
 
-class ParseError(ScriptError):
+class ScriptParseError(ScriptError):
     pass
 
 
-class EndOfFile(ParseError):
+class ScriptEndOfFile(ScriptParseError):
     pass
 
 
-class SyntaxError(ParseError):
+class ScriptSyntaxError(ScriptParseError):
     pass
 
 
-class UnknownFunction(ScriptError):
+class ScriptUnknownFunction(ScriptError):
     pass
 
 
-class ScriptText(unicode):
+class ScriptText(str):
 
     def eval(self, state):
         return self
 
 
+def normalize_tagname(name):
+    if name.startswith('_'):
+        return "~" + name[1:]
+    return name
+
+
 class ScriptVariable(object):
 
     def __init__(self, name):
@@ -64,10 +78,7 @@ class ScriptVariable(object):
         return '<ScriptVariable %%%s%%>' % self.name
 
     def eval(self, state):
-        name = self.name
-        if name.startswith(u"_"):
-            name = u"~" + name[1:]
-        return state.context.get(name, u"")
+        return state.context.get(normalize_tagname(self.name), "")
 
 
 FunctionRegistryItem = namedtuple("FunctionRegistryItem",
@@ -82,20 +93,25 @@ class ScriptFunction(object):
         try:
             argnum_bound = parser.functions[name].argcount
             argcount = len(args)
-            if argnum_bound and not (argnum_bound.lower <= argcount
-                                     and (argnum_bound.upper is None
-                                          or len(args) <= argnum_bound.upper)):
-                raise ScriptError(
-                    "Wrong number of arguments for $%s: Expected %s, got %i at position %i, line %i"
-                    % (name,
-                       str(argnum_bound.lower)
-                        if argnum_bound.upper is None
-                        else "%i - %i" % (argnum_bound.lower, argnum_bound.upper),
-                       argcount,
-                       parser._x,
-                       parser._y))
+            if argnum_bound:
+                too_few_args = argcount < argnum_bound.lower
+                if argnum_bound.upper is not None:
+                    if argnum_bound.lower == argnum_bound.upper:
+                        expected = "exactly %i" % argnum_bound.lower
+                    else:
+                        expected = "between %i and %i" % (argnum_bound.lower, argnum_bound.upper)
+                    too_many_args = argcount > argnum_bound.upper
+                else:
+                    expected = "at least %i" % argnum_bound.lower
+                    too_many_args = False
+
+                if too_few_args or too_many_args:
+                    raise ScriptError(
+                        "Wrong number of arguments for $%s: Expected %s, got %i at position %i, line %i"
+                        % (name, expected, argcount, parser._x, parser._y)
+                    )
         except KeyError:
-            raise UnknownFunction("Unknown function '%s'" % name)
+            raise ScriptUnknownFunction("Unknown function '%s'" % name)
 
         self.name = name
         self.args = args
@@ -104,7 +120,11 @@ class ScriptFunction(object):
         return "<ScriptFunction $%s(%r)>" % (self.name, self.args)
 
     def eval(self, parser):
-        function, eval_args, num_args = parser.functions[self.name]
+        try:
+            function, eval_args, num_args = parser.functions[self.name]
+        except KeyError:
+            raise ScriptUnknownFunction("Unknown function '%s'" % self.name)
+
         if eval_args:
             args = [arg.eval(parser) for arg in self.args]
         else:
@@ -115,10 +135,7 @@ class ScriptFunction(object):
 class ScriptExpression(list):
 
     def eval(self, state):
-        result = []
-        for item in self:
-            result.append(item.eval(state))
-        return "".join(result)
+        return "".join([item.eval(state) for item in self])
 
 
 def isidentif(ch):
@@ -127,7 +144,7 @@ def isidentif(ch):
 
 class ScriptParser(object):
 
-    """Tagger script parser.
+    r"""Tagger script parser.
 
 Grammar:
   text       ::= [^$%] | '\$' | '\%' | '\(' | '\)' | '\,'
@@ -139,17 +156,17 @@ Grammar:
   argument   ::= (variable | function | argtext)*
 """
 
-    _function_registry = ExtensionPoint()
+    _function_registry = ExtensionPoint(label='function_registry')
     _cache = {}
 
     def __raise_eof(self):
-        raise EndOfFile("Unexpected end of script at position %d, line %d" % (self._x, self._y))
+        raise ScriptEndOfFile("Unexpected end of script at position %d, line %d" % (self._x, self._y))
 
     def __raise_char(self, ch):
         #line = self._text[self._line:].split("\n", 1)[0]
         #cursor = " " * (self._pos - self._line - 1) + "^"
-        #raise SyntaxError("Unexpected character '%s' at position %d, line %d\n%s\n%s" % (ch, self._x, self._y, line, cursor))
-        raise SyntaxError("Unexpected character '%s' at position %d, line %d" % (ch, self._x, self._y))
+        #raise ScriptSyntaxError("Unexpected character '%s' at position %d, line %d\n%s\n%s" % (ch, self._x, self._y, line, cursor))
+        raise ScriptSyntaxError("Unexpected character '%s' at position %d, line %d" % (ch, self._x, self._y))
 
     def read(self):
         try:
@@ -192,7 +209,7 @@ Grammar:
             if ch == '(':
                 name = self._text[start:self._pos-1]
                 if name not in self.functions:
-                    raise UnknownFunction("Unknown function '%s'" % name)
+                    raise ScriptUnknownFunction("Unknown function '%s'" % name)
                 return ScriptFunction(name, self.parse_arguments(), self)
             elif ch is None:
                 self.__raise_eof()
@@ -282,6 +299,14 @@ Grammar:
         return ScriptParser._cache[key].eval(self)
 
 
+def enabled_tagger_scripts_texts():
+    """Returns an iterator over the enabled tagger scripts.
+    For each script, you'll get a tuple consisting of the script name and text"""
+    if not config.setting["enable_tagger_scripts"]:
+        return []
+    return [(s_name, s_text) for _s_pos, s_name, s_enabled, s_text in config.setting["list_of_scripts"] if s_enabled and s_text]
+
+
 def register_script_function(function, name=None, eval_args=True,
                              check_argcount=True):
     """Registers a script function. If ``name`` is ``None``,
@@ -291,7 +316,14 @@ def register_script_function(function, name=None, eval_args=True,
     If ``check_argcount`` is ``False`` the number of arguments passed to the
     function will not be verified."""
 
-    args, varargs, keywords, defaults = getargspec(function)
+    args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations = getfullargspec(function)
+
+    required_kwonlyargs = len(kwonlyargs)
+    if kwonlydefaults is not None:
+        required_kwonlyargs -= len(kwonlydefaults.keys())
+    if required_kwonlyargs:
+        raise TypeError("Functions with required keyword-only parameters are not supported")
+
     args = len(args) - 1  # -1 for the parser
     varargs = varargs is not None
     defaults = len(defaults) if defaults else 0
@@ -302,12 +334,36 @@ def register_script_function(function, name=None, eval_args=True,
         name = function.__name__
     ScriptParser._function_registry.register(function.__module__,
         (name, FunctionRegistryItem(
-                    function, eval_args,
-                    argcount if argcount and check_argcount else False)
+            function, eval_args,
+            argcount if argcount and check_argcount else False)
          )
     )
 
 
+def script_function(name=None, eval_args=True, check_argcount=True, prefix='func_'):
+    """Decorator helper to register script functions
+
+    It calls ``register_script_function()`` and share same arguments
+    Extra optional arguments:
+        ``prefix``: define the prefix to be removed from defined function to name script function
+                    By default, ``func_foo`` will create ``foo`` script function
+
+    Example:
+        @script_function(eval_args=False)
+        def func_myscriptfunc():
+            ...
+    """
+    def script_function_decorator(func):
+        fname = func.__name__
+        if name is None and prefix and fname.startswith(prefix):
+            sname = fname[len(prefix):]
+        else:
+            sname = name
+        register_script_function(func, name=sname, eval_args=eval_args, check_argcount=check_argcount)
+        return func
+    return script_function_decorator
+
+
 def _compute_int(operation, *args):
     return str(reduce(operation, map(int, args)))
 
@@ -316,6 +372,28 @@ def _compute_logic(operation, *args):
     return operation(args)
 
 
+def _get_multi_values(parser, multi, separator):
+    if isinstance(separator, ScriptExpression):
+        separator = separator.eval(parser)
+
+    if separator == MULTI_VALUED_JOINER:
+        # Convert ScriptExpression containing only a single variable into variable
+        if (isinstance(multi, ScriptExpression)
+            and len(multi) == 1
+            and isinstance(multi[0], ScriptVariable)):
+            multi = multi[0]
+
+        # If a variable, return multi-values
+        if isinstance(multi, ScriptVariable):
+            return parser.context.getall(normalize_tagname(multi.name))
+
+    # Fall-back to converting to a string and splitting if haystack is an expression
+    # or user has overridden the separator character.
+    multi = multi.eval(parser)
+    return multi.split(separator) if separator else [multi]
+
+
+@script_function(eval_args=False)
 def func_if(parser, _if, _then, _else=None):
     """If ``if`` is not empty, it returns ``then``, otherwise it returns ``else``."""
     if _if.eval(parser):
@@ -325,6 +403,7 @@ def func_if(parser, _if, _then, _else=None):
     return ''
 
 
+@script_function(eval_args=False)
 def func_if2(parser, *args):
     """Returns first non empty argument."""
     for arg in args:
@@ -334,11 +413,13 @@ def func_if2(parser, *args):
     return ''
 
 
+@script_function(eval_args=False)
 def func_noop(parser, *args):
     """Does nothing :)"""
     return ''
 
 
+@script_function()
 def func_left(parser, text, length):
     """Returns first ``num`` characters from ``text``."""
     try:
@@ -347,6 +428,7 @@ def func_left(parser, text, length):
         return ""
 
 
+@script_function()
 def func_right(parser, text, length):
     """Returns last ``num`` characters from ``text``."""
     try:
@@ -355,16 +437,19 @@ def func_right(parser, text, length):
         return ""
 
 
+@script_function()
 def func_lower(parser, text):
     """Returns ``text`` in lower case."""
     return text.lower()
 
 
+@script_function()
 def func_upper(parser, text):
     """Returns ``text`` in upper case."""
     return text.upper()
 
 
+@script_function()
 def func_pad(parser, text, length, char):
     try:
         return char * (int(length) - len(text)) + text
@@ -372,14 +457,17 @@ def func_pad(parser, text, length, char):
         return ""
 
 
+@script_function()
 def func_strip(parser, text):
-    return re.sub("\s+", " ", text).strip()
+    return re.sub(r"\s+", " ", text).strip()
 
 
+@script_function()
 def func_replace(parser, text, old, new):
     return text.replace(old, new)
 
 
+@script_function()
 def func_in(parser, text, needle):
     if needle in text:
         return "1"
@@ -387,101 +475,125 @@ def func_in(parser, text, needle):
         return ""
 
 
-def func_inmulti(parser, text, value, separator=MULTI_VALUED_JOINER):
-    """Splits ``text`` by ``separator``, and returns true if the resulting list contains ``value``."""
-    return func_in(parser, text.split(separator) if separator else [text], value)
+@script_function(eval_args=False)
+def func_inmulti(parser, haystack, needle, separator=MULTI_VALUED_JOINER):
+    """Searches for ``needle`` in ``haystack``, supporting a list variable for
+       ``haystack``. If a string is used instead, then a ``separator`` can be
+       used to split it. In both cases, it returns true if the resulting list
+       contains exactly ``needle`` as a member."""
+
+    needle = needle.eval(parser)
+    return func_in(parser, _get_multi_values(parser, haystack, separator), needle)
 
 
+@script_function()
 def func_rreplace(parser, text, old, new):
-    return re.sub(old, new, text)
+    try:
+        return re.sub(old, new, text)
+    except re.error:
+        return text
 
 
+@script_function()
 def func_rsearch(parser, text, pattern):
-    match = re.search(pattern, text)
+    try:
+        match = re.search(pattern, text)
+    except re.error:
+        return ""
     if match:
         try:
             return match.group(1)
         except IndexError:
             return match.group(0)
-    return u""
+    return ""
 
 
+@script_function()
 def func_num(parser, text, length):
     try:
-        format = "%%0%dd" % min(int(length), 20)
+        format_ = "%%0%dd" % min(int(length), 20)
     except ValueError:
         return ""
     try:
         value = int(text)
     except ValueError:
         value = 0
-    return format % value
+    return format_ % value
 
 
+@script_function()
 def func_unset(parser, name):
     """Unsets the variable ``name``."""
-    if name.startswith("_"):
-        name = "~" + name[1:]
+    name = normalize_tagname(name)
     # Allow wild-card unset for certain keys
     if name in ('performer:*', 'comment:*', 'lyrics:*'):
         name = name[:-1]
-        for key in parser.context.keys():
+        for key in list(parser.context.keys()):
             if key.startswith(name):
-                del parser.context[key]
+                parser.context.unset(key)
         return ""
     try:
-        del parser.context[name]
+        parser.context.unset(name)
     except KeyError:
         pass
     return ""
 
 
+@script_function()
+def func_delete(parser, name):
+    """
+    Deletes the variable ``name``.
+    This will unset the tag with the given name and also mark the tag for
+    deletion on save.
+    """
+    parser.context.delete(normalize_tagname(name))
+    return ""
+
+
+@script_function()
 def func_set(parser, name, value):
     """Sets the variable ``name`` to ``value``."""
     if value:
-        if name.startswith("_"):
-            name = "~" + name[1:]
-        parser.context[name] = value
+        parser.context[normalize_tagname(name)] = value
     else:
         func_unset(parser, name)
     return ""
 
 
+@script_function()
 def func_setmulti(parser, name, value, separator=MULTI_VALUED_JOINER):
     """Sets the variable ``name`` to ``value`` as a list; splitting by the passed string, or "; " otherwise."""
     return func_set(parser, name, value.split(separator) if value and separator else value)
 
 
+@script_function()
 def func_get(parser, name):
     """Returns the variable ``name`` (equivalent to ``%name%``)."""
-    if name.startswith("_"):
-        name = "~" + name[1:]
-    return parser.context.get(name, u"")
+    return parser.context.get(normalize_tagname(name), "")
 
 
+@script_function()
 def func_copy(parser, new, old):
     """Copies content of variable ``old`` to variable ``new``."""
-    if new.startswith("_"):
-        new = "~" + new[1:]
-    if old.startswith("_"):
-        old = "~" + old[1:]
+    new = normalize_tagname(new)
+    old = normalize_tagname(old)
     parser.context[new] = parser.context.getall(old)[:]
     return ""
 
 
+@script_function()
 def func_copymerge(parser, new, old):
     """Copies content of variable ``old`` and appends it into variable ``new``, removing duplicates. This is normally
     used to merge a multi-valued variable into another, existing multi-valued variable."""
-    if new.startswith("_"):
-        new = "~" + new[1:]
-    if old.startswith("_"):
-        old = "~" + old[1:]
+    new = normalize_tagname(new)
+    old = normalize_tagname(old)
     newvals = parser.context.getall(new)
     oldvals = parser.context.getall(old)
-    parser.context[new] = newvals + list(set(oldvals) - set(newvals))
+    parser.context[new] = uniqify(newvals + oldvals)
     return ""
 
 
+@script_function()
 def func_trim(parser, text, char=None):
     """Trims all leading and trailing whitespaces from ``text``. The optional
        second parameter specifies the character to trim."""
@@ -491,6 +603,7 @@ def func_trim(parser, text, char=None):
         return text.strip()
 
 
+@script_function()
 def func_add(parser, x, y, *args):
     """Adds ``y`` to ``x``.
        Can be used with an arbitrary number of arguments.
@@ -502,6 +615,7 @@ def func_add(parser, x, y, *args):
         return ""
 
 
+@script_function()
 def func_sub(parser, x, y, *args):
     """Subtracts ``y`` from ``x``.
        Can be used with an arbitrary number of arguments.
@@ -513,17 +627,19 @@ def func_sub(parser, x, y, *args):
         return ""
 
 
+@script_function()
 def func_div(parser, x, y, *args):
     """Divides ``x`` by ``y``.
        Can be used with an arbitrary number of arguments.
        Eg: $div(x, y, z) = ((x / y) / z)
     """
     try:
-        return _compute_int(operator.div, x, y, *args)
+        return _compute_int(operator.floordiv, x, y, *args)
     except ValueError:
         return ""
 
 
+@script_function()
 def func_mod(parser, x, y, *args):
     """Returns the remainder of ``x`` divided by ``y``.
        Can be used with an arbitrary number of arguments.
@@ -535,6 +651,7 @@ def func_mod(parser, x, y, *args):
         return ""
 
 
+@script_function()
 def func_mul(parser, x, y, *args):
     """Multiplies ``x`` by ``y``.
        Can be used with an arbitrary number of arguments.
@@ -546,6 +663,7 @@ def func_mul(parser, x, y, *args):
         return ""
 
 
+@script_function()
 def func_or(parser, x, y, *args):
     """Returns true, if either ``x`` or ``y`` not empty.
        Can be used with an arbitrary number of arguments. The result is
@@ -557,6 +675,7 @@ def func_or(parser, x, y, *args):
         return ""
 
 
+@script_function()
 def func_and(parser, x, y, *args):
     """Returns true, if both ``x`` and ``y`` are not empty.
        Can be used with an arbitrary number of arguments. The result is
@@ -568,6 +687,7 @@ def func_and(parser, x, y, *args):
         return ""
 
 
+@script_function()
 def func_not(parser, x):
     """Returns true, if ``x`` is empty."""
     if not x:
@@ -576,6 +696,7 @@ def func_not(parser, x):
         return ""
 
 
+@script_function()
 def func_eq(parser, x, y):
     """Returns true, if ``x`` equals ``y``."""
     if x == y:
@@ -584,6 +705,7 @@ def func_eq(parser, x, y):
         return ""
 
 
+@script_function()
 def func_ne(parser, x, y):
     """Returns true, if ``x`` not equals ``y``."""
     if x != y:
@@ -592,6 +714,7 @@ def func_ne(parser, x, y):
         return ""
 
 
+@script_function()
 def func_lt(parser, x, y):
     """Returns true, if ``x`` is lower than ``y``."""
     try:
@@ -602,6 +725,7 @@ def func_lt(parser, x, y):
     return ""
 
 
+@script_function()
 def func_lte(parser, x, y):
     """Returns true, if ``x`` is lower than or equals ``y``."""
     try:
@@ -612,6 +736,7 @@ def func_lte(parser, x, y):
     return ""
 
 
+@script_function()
 def func_gt(parser, x, y):
     """Returns true, if ``x`` is greater than ``y``."""
     try:
@@ -622,6 +747,7 @@ def func_gt(parser, x, y):
     return ""
 
 
+@script_function()
 def func_gte(parser, x, y):
     """Returns true, if ``x`` is greater than or equals ``y``."""
     try:
@@ -632,10 +758,17 @@ def func_gte(parser, x, y):
     return ""
 
 
+@script_function()
 def func_len(parser, text=""):
     return str(len(text))
 
 
+@script_function(eval_args=False)
+def func_lenmulti(parser, multi, separator=MULTI_VALUED_JOINER):
+    return func_len(parser, _get_multi_values(parser, multi, separator))
+
+
+@script_function()
 def func_performer(parser, pattern="", join=", "):
     values = []
     for name, value in parser.context.items():
@@ -644,19 +777,26 @@ def func_performer(parser, pattern="", join=", "):
     return join.join(values)
 
 
-def func_matchedtracks(parser, arg):
-    if parser.file and parser.file.parent:
+@script_function(eval_args=False)
+def func_matchedtracks(parser, *args):
+    # only works in file naming scripts, always returns zero in tagging scripts
+    file = parser.file
+    if file and file.parent and hasattr(file.parent, 'album'):
         return str(parser.file.parent.album.get_num_matched_tracks())
     return "0"
 
 
+@script_function()
 def func_is_complete(parser):
-    if (parser.file and parser.file.parent
-        and parser.file.parent.album.is_complete()):
+    # only works in file naming scripts, always returns zero in tagging scripts
+    file = parser.file
+    if (file and file.parent and hasattr(file.parent, 'album')
+            and file.parent.album.is_complete()):
         return "1"
-    return "0"
+    return ""
 
 
+@script_function()
 def func_firstalphachar(parser, text="", nonalpha="#"):
     if len(text) == 0:
         return nonalpha
@@ -667,14 +807,16 @@ def func_firstalphachar(parser, text="", nonalpha="#"):
         return nonalpha
 
 
+@script_function()
 def func_initials(parser, text=""):
     return "".join(a[:1] for a in text.split(" ") if a[:1].isalpha())
 
 
+@script_function()
 def func_firstwords(parser, text, length):
     try:
         length = int(length)
-    except ValueError as e:
+    except ValueError:
         length = 0
     if len(text) <= length:
         return text
@@ -684,26 +826,30 @@ def func_firstwords(parser, text, length):
         return text[:length].rsplit(' ', 1)[0]
 
 
+@script_function()
 def func_startswith(parser, text, prefix):
     if text.startswith(prefix):
         return "1"
-    return "0"
+    return ""
 
 
+@script_function()
 def func_endswith(parser, text, suffix):
     if text.endswith(suffix):
         return "1"
-    return "0"
+    return ""
 
 
+@script_function()
 def func_truncate(parser, text, length):
     try:
         length = int(length)
-    except ValueError as e:
+    except ValueError:
         length = None
     return text[:length].rstrip()
 
 
+@script_function(check_argcount=False)
 def func_swapprefix(parser, text, *prefixes):
     """
     Moves the specified prefixes to the end of text.
@@ -717,6 +863,7 @@ def func_swapprefix(parser, text, *prefixes):
     return text
 
 
+@script_function(check_argcount=False)
 def func_delprefix(parser, text, *prefixes):
     """
     Deletes the specified prefixes.
@@ -738,13 +885,15 @@ def _delete_prefix(parser, text, *prefixes):
     if not prefixes:
         prefixes = ('A', 'The')
     text = text.strip()
-    match = re.match('(' + r'\s+)|('.join(prefixes) + r'\s+)', text)
+    rx = '(' + r'\s+)|('.join(map(re.escape, prefixes)) + r'\s+)'
+    match = re.match(rx, text)
     if match:
         pref = match.group()
         return text[len(pref):], pref.strip()
     return text, ''
 
 
+@script_function(check_argcount=False)
 def func_eq_any(parser, x, *args):
     """
     Return True if one string matches any of one or more other strings.
@@ -755,6 +904,7 @@ def func_eq_any(parser, x, *args):
     return '1' if x in args else ''
 
 
+@script_function(check_argcount=False)
 def func_ne_all(parser, x, *args):
     """
     Return True if one string doesn't match all of one or more other strings.
@@ -765,6 +915,7 @@ def func_ne_all(parser, x, *args):
     return '1' if x not in args else ''
 
 
+@script_function(check_argcount=False)
 def func_eq_all(parser, x, *args):
     """
     Return True if all string are equal.
@@ -777,6 +928,7 @@ def func_eq_all(parser, x, *args):
     return '1'
 
 
+@script_function(check_argcount=False)
 def func_ne_any(parser, x, *args):
     """
     Return True if all strings are not equal.
@@ -786,55 +938,289 @@ def func_ne_any(parser, x, *args):
     return func_not(parser, func_eq_all(parser, x, *args))
 
 
-register_script_function(func_if, "if", eval_args=False)
-register_script_function(func_if2, "if2", eval_args=False)
-register_script_function(func_noop, "noop", eval_args=False)
-register_script_function(func_left, "left")
-register_script_function(func_right, "right")
-register_script_function(func_lower, "lower")
-register_script_function(func_upper, "upper")
-register_script_function(func_pad, "pad")
-register_script_function(func_strip, "strip")
-register_script_function(func_replace, "replace")
-register_script_function(func_rreplace, "rreplace")
-register_script_function(func_rsearch, "rsearch")
-register_script_function(func_num, "num")
-register_script_function(func_unset, "unset")
-register_script_function(func_set, "set")
-register_script_function(func_setmulti, "setmulti")
-register_script_function(func_get, "get")
-register_script_function(func_trim, "trim")
-register_script_function(func_add, "add")
-register_script_function(func_sub, "sub")
-register_script_function(func_div, "div")
-register_script_function(func_mod, "mod")
-register_script_function(func_mul, "mul")
-register_script_function(func_or, "or")
-register_script_function(func_and, "and")
-register_script_function(func_not, "not")
-register_script_function(func_eq, "eq")
-register_script_function(func_ne, "ne")
-register_script_function(func_lt, "lt")
-register_script_function(func_lte, "lte")
-register_script_function(func_gt, "gt")
-register_script_function(func_gte, "gte")
-register_script_function(func_in, "in")
-register_script_function(func_inmulti, "inmulti")
-register_script_function(func_copy, "copy")
-register_script_function(func_copymerge, "copymerge")
-register_script_function(func_len, "len")
-register_script_function(func_performer, "performer")
-register_script_function(func_matchedtracks, "matchedtracks")
-register_script_function(func_is_complete, "is_complete")
-register_script_function(func_firstalphachar, "firstalphachar")
-register_script_function(func_initials, "initials")
-register_script_function(func_firstwords, "firstwords")
-register_script_function(func_startswith, "startswith")
-register_script_function(func_endswith, "endswith")
-register_script_function(func_truncate, "truncate")
-register_script_function(func_swapprefix, "swapprefix", check_argcount=False)
-register_script_function(func_delprefix, "delprefix", check_argcount=False)
-register_script_function(func_eq_any, "eq_any", check_argcount=False)
-register_script_function(func_ne_all, "ne_all", check_argcount=False)
-register_script_function(func_eq_all, "eq_all", check_argcount=False)
-register_script_function(func_ne_any, "ne_any", check_argcount=False)
+@script_function()
+def func_title(parser, text):
+    # GPL 2.0 licensed code by Javier Kohen, Sambhav Kothari
+    # from https://github.com/metabrainz/picard-plugins/blob/2.0/plugins/titlecase/titlecase.py
+    """
+    Title-case a text - capitalizes first letter of every word
+    like: from "Lost in the Supermarket" to "Lost In The Supermarket"
+    Example: $set(album,$title(%album%))
+    """
+    if not text:
+        return ""
+    capitalized = text[0].capitalize()
+    capital = False
+    for i in range(1, len(text)):
+        t = text[i]
+        if t in "’'" and text[i-1].isalpha():
+            capital = False
+        elif iswbound(t):
+            capital = True
+        elif capital and t.isalpha():
+            capital = False
+            t = t.capitalize()
+        else:
+            capital = False
+        capitalized += t
+    return capitalized
+
+
+def iswbound(char):
+    # GPL 2.0 licensed code by Javier Kohen, Sambhav Kothari
+    # from https://github.com/metabrainz/picard-plugins/blob/2.0/plugins/titlecase/titlecase.py
+    """ Checks whether the given character is a word boundary """
+    category = unicodedata.category(char)
+    return "Zs" == category or "Sk" == category or "P" == category[0]
+
+
+@script_function()
+def func_is_audio(parser):
+    """Returns true, if the file processed is an audio file."""
+    if func_is_video(parser) == "1":
+        return ""
+    else:
+        return "1"
+
+
+@script_function()
+def func_is_video(parser):
+    """Returns true, if the file processed is a video file."""
+    if parser.context['~video'] and parser.context['~video'] != '0':
+        return "1"
+    else:
+        return ""
+
+
+@script_function()
+def func_find(parser, haystack, needle):
+    """Find the location of the first occurrence of one string within another.
+
+    Arguments:
+        parser: The ScriptParser object used to parse the script.
+        haystack: The string to search.
+        needle: The substring to find.
+
+    Returns:
+        The zero-based index of the first occurrance of needle in haystack, or -1 if needle was not found.
+    """
+    return str(haystack.find(needle))
+
+
+@script_function()
+def func_reverse(parser, text):
+    """Returns 'text' in reverse order.
+
+    Arguments:
+        parser: The ScriptParser object used to parse the script.
+        text: String to be processed.
+
+    Returns:
+        Text in reverse order.
+    """
+    return text[::-1]
+
+
+@script_function()
+def func_substr(parser, text, start_index, end_index):
+    """Extract a specified portion of a string.
+
+    Arguments:
+        parser: The ScriptParser object used to parse the script.
+        text: The string from which the extract will be made.
+        start_index: Integer index of the first character to extract.
+        end_index: Integer index of the first character that will not be extracted.
+
+    Returns:
+        Returns the substring beginning with the character at the start index,
+        up to (but not including) the character at the end index.  The first
+        character is at index number 0.  If the start index is left blank, it
+        defaults to the first character in the string.  If the end index is
+        left blank, it defaults to the number of characters in the string.
+        If either index is negative, it is subtracted from the total number of
+        characters in the string to provide the index used.
+    """
+    try:
+        start = int(start_index) if start_index else None
+    except ValueError:
+        start = None
+    try:
+        end = int(end_index) if end_index else None
+    except ValueError:
+        end = None
+    return text[start:end]
+
+
+@script_function(eval_args=False)
+def func_getmulti(parser, multi, item_index, separator=MULTI_VALUED_JOINER):
+    """Returns value of the item at the specified index in the multi-value variable.  Index values are zero-based.
+
+    Arguments:
+        parser: The ScriptParser object used to parse the script.
+        multi: The multi-value from which the item is to be retrieved.
+        item_index: The zero-based integer index of the item to be retrieved.
+        separator: String used to separate the elements in the multi-value.
+
+    Returns:
+        Returns the value of the item at the specified index in the multi-value variable.
+    """
+    if not item_index:
+        return ''
+    try:
+        index = int(item_index.eval(parser))
+        multi_var = _get_multi_values(parser, multi, separator)
+        return str(multi_var[index])
+    except (ValueError, IndexError):
+        return ''
+
+
+@script_function(eval_args=False)
+def func_foreach(parser, multi, loop_code, separator=MULTI_VALUED_JOINER):
+    """Iterates over each element found in the specified multi-value variable.
+
+    Iterates over each element found in the specified multi-value variable, executing the specified code.
+    For each loop, the element value is first stored in the tag specified by _loop_value and the count is
+    stored in the tag specified by _loop_count.  This allows the element or count value to be accessed within
+    the code script.
+
+    Arguments:
+        parser: The ScriptParser object used to parse the script.
+        multi: The multi-value to be iterated.
+        loop_code: String of script code to be processed on each iteration.
+        separator: String used to separate the elements in the multi-value.
+    """
+    multi_value = _get_multi_values(parser, multi, separator)
+    for loop_count, value in enumerate(multi_value, 1):
+        func_set(parser, '_loop_count', str(loop_count))
+        func_set(parser, '_loop_value', str(value))
+        loop_code.eval(parser)
+    func_unset(parser, '_loop_count')
+    func_unset(parser, '_loop_value')
+    return ''
+
+
+@script_function(eval_args=False)
+def func_while(parser, condition, loop_code):
+    """Standard 'while' loop.  Also includes a runaway check to limit the maximum number of iterations.
+
+    Arguments:
+        parser: The ScriptParser object used to parse the script.
+        condition: String of script code to check before each iteration through the loop.
+        loop_code: String of script code to be processed on each iteration.
+    """
+    if condition and loop_code:
+        runaway_check = 1000
+        loop_count = 0
+        while condition.eval(parser) and loop_count < runaway_check:
+            loop_count += 1
+            func_set(parser, '_loop_count', str(loop_count))
+            loop_code.eval(parser)
+        func_unset(parser, '_loop_count')
+    return ''
+
+
+@script_function(eval_args=False)
+def func_map(parser, multi, loop_code, separator=MULTI_VALUED_JOINER):
+    """Iterates over each element found in the specified multi-value variable and updates the value.
+
+    Iterates over each element found in the specified multi-value variable and updates the value of the
+    element to the value returned by the specified code. For each loop, the element value is first stored in
+    the tag specified by _loop_value and the count is stored in the tag specified by _loop_count.  This
+    allows the element or count value to be accessed within the code script.
+
+    Arguments:
+        parser: The ScriptParser object used to parse the script.
+        multi: The multi-value to be iterated.
+        loop_code: String of script code to be processed on each iteration that yields the new value for
+            the multi-value element.
+        separator: String used to separate the elements in the multi-value.
+
+    Returns the updated multi-value variable.
+    """
+    multi_value = _get_multi_values(parser, multi, separator)
+    for loop_count, value in enumerate(multi_value, 1):
+        func_set(parser, '_loop_count', str(loop_count))
+        func_set(parser, '_loop_value', str(value))
+        multi_value[loop_count - 1] = str(loop_code.eval(parser))
+    func_unset(parser, '_loop_count')
+    func_unset(parser, '_loop_value')
+    if not isinstance(separator, str):
+        separator = separator.eval(parser)
+    return separator.join(multi_value)
+
+
+@script_function(eval_args=False)
+def func_join(parser, multi, join_phrase, separator=MULTI_VALUED_JOINER):
+    """Joins all elements in the specified multi-value variable, placing the join_phrase between each element.
+
+    Arguments:
+        parser: The ScriptParser object used to parse the script.
+        multi: The ScriptVariable/Function that evaluates to a multi-value whose
+            elements are to be joined.
+        join_phrase: The ScriptVariable/Function that evaluates to a string which
+            will be placed between each of the elements.
+        separator: A string or the ScriptVariable/Function that evaluates to the
+            string used to separate the elements in the multi-value.
+
+    Returns a string with the elements joined.
+    """
+    join_phrase = str(join_phrase.eval(parser))
+    multi_value = _get_multi_values(parser, multi, separator)
+    return join_phrase.join(multi_value)
+
+
+@script_function(eval_args=False)
+def func_slice(parser, multi, start_index, end_index, separator=MULTI_VALUED_JOINER):
+    """Returns a multi-value containing a slice of the supplied multi-value.  Index values are zero-based.
+
+    Arguments:
+        parser: The ScriptParser object used to parse the script.
+        multi: The ScriptVariable/Function that evaluates to a multi-value from
+            which the slice is to be retrieved.
+        start_index: The ScriptVariable/Function that evaluates to a zero-based integer
+            index of the first item included in the slice.
+        end_index: The ScriptVariable/Function that evaluates to a zero-based integer
+            index of the first item not included in the slice.
+        separator: A string or the ScriptVariable/Function that evaluates to the
+            string used to separate the elements in the multi-value.
+
+    Returns:
+        Returns a multi-value variable containing the specified slice.
+    """
+    try:
+        start = int(start_index.eval(parser)) if start_index else None
+    except ValueError:
+        start = None
+    try:
+        end = int(end_index.eval(parser)) if end_index else None
+    except ValueError:
+        end = None
+    multi_var = _get_multi_values(parser, multi, separator)
+    if not isinstance(separator, str):
+        separator = separator.eval(parser)
+    return separator.join(multi_var[start:end])
+
+
+@script_function()
+def func_datetime(parser, format=None):
+    """Return the current date and time as a string.
+
+    Arguments:
+        parser: The ScriptParser object used to parse the script.
+        format: A string or the ScriptVariable/Function that evaluates to the
+            string used to format the output.  Default is '%Y-%m-%d %H:%M:%S'
+            if blank.  Uses strftime() format.
+
+    Returns:
+        Returns the current date and time as a string.
+    """
+    # local_tz required for Python 3.5 which does not allow setting astimezone()
+    # on a naive datetime.datetime object.  This provides timezone information to
+    # allow the use of %Z and %z in the output format.
+    local_tz = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
+
+    # Handle case where format evaluates to ''
+    if not format:
+        format = '%Y-%m-%d %H:%M:%S'
+
+    return datetime.datetime.now(tz=local_tz).strftime(format)
