@@ -44,8 +44,15 @@ class OptionsPage(QtWidgets.QWidget):
         super().__init__(*args, **kwargs)
         self.setStyleSheet(self.STYLESHEET)
 
-    def info(self):
-        raise NotImplementedError
+        # Keep track whether the options page has been destroyed to avoid
+        # trying to update deleted UI widgets after plugin list refresh.
+        self.deleted = False
+
+        # The on destroyed cannot be created as a method on this class or it will never get called.
+        # See https://stackoverflow.com/questions/16842955/widgets-destroyed-signal-is-not-fired-pyqt
+        def on_destroyed(obj=None):
+            self.deleted = True
+        self.destroyed.connect(on_destroyed)
 
     def check(self):
         pass
@@ -101,7 +108,7 @@ class OptionsPage(QtWidgets.QWidget):
         regex_edit.textChanged.connect(live_checker)
 
 
-_pages = ExtensionPoint()
+_pages = ExtensionPoint(label='pages')
 
 
 def register_options_page(page_class):
